﻿using System;
using System.Diagnostics;
using System.Collections.Generic;

namespace CrossWordCSharp
{
    class IndexTree
    {
        int w, h;
        int[][,] T;
        int levels;
        int M;
        public IndexTree(int width, int height)
        {
            w = width;
            h = height;
            levels = 0;
            while ((1 << levels) < Math.Max(w, h)) levels++;
            M = 1 << levels;
            T = new int[levels + 1][,];
            // T[0] = root level ([1x1], e.g., represents a single 512x512 square).
            // in the same spirit T[1] = one level directly below; [2x2], represents four 256x256 squares
            for (int i = 0; i <= levels; i++) {
                T[i] = new int[1 << i, 1 << i];
            }
            for (int y = 0; y < M; y++)
                for (int x = 0; x < M; x++)
                    if (x >= w || y >= h)
                        update(x, y, 1); // mask off the end of the initial area.

        }
        int free() { return M * M - T[0][0, 0]; }
        int free(int L, int x, int y)
        {
            return M * M / (1 << (2 * L)) - T[L][y, x];
        }
        public double getFillRate() { return 1.0 - (double)free() / (double) (w * h); }
        // modify a cell. Amount can be +1 or -1. If a cell is already set, and is +1'd again, nothing happens.
        public void update(int x, int y, int amount)
        {
            if (amount == 0 || (amount == 1 && T[levels][y, x] == 1)) return;
            if (amount == -1) Debug.Assert(T[levels][y, x] == 1);
            for (int i = levels; i >= 0; i--) {
                T[i][y, x] += amount;
                x /= 2;
                y /= 2;
            }
        }
        static int manhattan(int x1, int y1, int x2, int y2)
        {
            return Math.Abs(x1 - x2) + Math.Abs(y1 - y2);
        }
        int evalScore(int L, int zx, int zy, int tx, int ty)
        {
            int CS = (1 << (levels - L));
            int x0 = zx * CS, x1 = (zx + 1) * CS, y0 = zy * CS, y1 = (zy + 1) * CS;
            int fx = Math.Min(x1 - 1, Math.Max(tx, x0));
            int fy = Math.Min(y1 - 1, Math.Max(ty, y0));
            return manhattan(fx, fy, tx, ty);
        }
        // returns a 3-tuple (X, Y, score)
        Tuple<int, int, int> findRecurse(int tx, int ty, int L, int zx, int zy)
        {
            if (L == levels) {
                Debug.Assert(T[L][zy, zx] == 0);
                return new Tuple<int, int, int>(zx, zy, manhattan(tx, ty, zx, zy));
            } else {
                var l = new List<Tuple<int, int, int>>();
                for (int dy = 0; dy < 2; dy++)
                    for (int dx = 0; dx < 2; dx++) 
                        if (free(L + 1, zx * 2 + dx, zy * 2 + dy) > 0)
                            l.Add(new Tuple<int, int, int>(evalScore(L + 1, zx * 2 + dx, zy * 2 + dy, tx, ty), zy * 2 + dy, zx * 2 + dx));
                l.Sort();
                int bestScore = 999999;
                var candidate = new Tuple<int, int, int>(0, 0, -1);
                foreach (Tuple<int, int, int> tp in l) {
                    if (free(L + 1, tp.Item3, tp.Item2) <= 0) continue;
                    if (tp.Item1 > bestScore) break;
                    var temp = findRecurse(tx, ty, L + 1, tp.Item3, tp.Item2);
                    if (temp.Item3 < bestScore) {
                        bestScore = temp.Item3;
                        candidate = temp;
                    }
                }
                Debug.Assert(candidate.Item3 != -1);
                return candidate;
            }
        }
        // find the closest free cell to a specified target (using manhattan metric)
        public Tuple<int, int> findClosestFreeCell(int tx, int ty)
        {
            var temp = findRecurse(tx, ty, 0, 0, 0);
            return Tuple.Create(temp.Item1, temp.Item2);
        }
    }
}
